package com.indiscale.fdo.manager.service.fdo;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;

import com.indiscale.fdo.DefaultData;
import com.indiscale.fdo.DefaultMetadat;
import com.indiscale.fdo.api.DigitalObject;
import com.indiscale.fdo.api.FDO;
import com.indiscale.fdo.api.FdoProfile;
import com.indiscale.fdo.api.InputStreamSource;
import com.indiscale.fdo.api.Manager;
import com.indiscale.fdo.api.PidUnresolvableException;
import com.indiscale.fdo.api.RepositoryConnection;
import com.indiscale.fdo.api.TokenAuthenticationInfo;
import com.indiscale.fdo.api.UnknownRepositoryException;
import com.indiscale.fdo.api.ValidationException;
import com.indiscale.fdo.manager.service.BaseController;
import com.indiscale.fdo.manager.service.api.model.Links;
import com.indiscale.fdo.manager.service.api.model.ResolvePID200Response;
import com.indiscale.fdo.manager.service.api.model.TargetRepositories;
import com.indiscale.fdo.manager.service.api.operation.FdoApi;
import jakarta.validation.Valid;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.CrossOrigin;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.server.ResponseStatusException;

class MultipartFileWrapper implements InputStreamSource {

  private MultipartFile wrapped;

  public MultipartFileWrapper(MultipartFile file) {
    this.wrapped = file;
  }

  @Override
  public InputStream getInputStream() throws IOException {
    return wrapped.getInputStream();
  }
}

class MetadataWrapper extends DefaultMetadat {

  private MultipartFile file;

  public MetadataWrapper(MultipartFile file) {
    super();
    this.file = file;
  }

  @Override
  public InputStream getInputStream() throws IOException {
    return this.file.getInputStream();
  }
}

class DataWrapper extends DefaultData {

  private MultipartFile file;

  public DataWrapper(MultipartFile file) {
    super();
    this.file = file;
  }

  @Override
  public InputStream getInputStream() throws IOException {
    return file.getInputStream();
  }
}

@RestController
@CrossOrigin(
    origins = {"${react-dev-server}"},
    exposedHeaders = {"Location"})
public class FDOApiImpl extends BaseController implements FdoApi {

  @Override
  public ResponseEntity<Void> createFDO(
      @Valid TargetRepositories repositories, MultipartFile data, MultipartFile metadata) {
    try (Manager manager = getManager()) {
      RepositoryConnection repository =
          manager.getRepositoryRegistry().createRepositoryConnection(repositories.getFdo());
      repository.setTokenAuthenticationInfo(new TokenAuthenticationInfo(getToken()));
      FdoProfile profile = manager.getDefaultProfile();
      FDO fdo =
          manager.createFDO(
              profile, repository, new DataWrapper(data), new MetadataWrapper(metadata));
      getLogger().createFDO(fdo, repository);

      // TODO(tf) this is a workaround for the apache proxy
      String[] pid = fdo.getPID().split("/", 2);

      URI location = linkTo(getOperations(FDOApiImpl.class).resolvePID(pid[0], pid[1])).toUri();
      return ResponseEntity.created(location).build();
    } catch (UnknownRepositoryException e) {
      throw new ResponseStatusException(HttpStatus.NOT_FOUND, "Not Found. Unknown repository id.");
    } catch (ValidationException e) {
      throw new ResponseStatusException(HttpStatus.BAD_REQUEST, e.getValidationResult().toString());
    }
  }

  @Override
  public ResponseEntity<ResolvePID200Response> resolvePID(String prefix, String suffix) {
    String pid = prefix + "/" + suffix;
    try (Manager manager = getManager()) {
      DigitalObject resolved = manager.resolvePID(pid);
      return ResponseEntity.ok(createResponse(resolved));
    } catch (PidUnresolvableException e) {
      throw new ResponseStatusException(
          HttpStatus.NOT_FOUND, "Not found. Could not resolve PID " + e.getPid());
    }
  }

  private ResolvePID200Response createResponse(DigitalObject resolved) {
    String[] pid = resolved.getPID().split("/", 2);
    com.indiscale.fdo.manager.service.api.model.DigitalObject data =
        new com.indiscale.fdo.manager.service.api.model.DigitalObject(
            resolved.getPID(), resolved.isFDO());
    Links self = linkSelf(getOperations(getClass()).resolvePID(pid[0], pid[1]));

    return new ResolvePID200Response().data(data).links(self);
  }
}
